#ifndef DAQCHANNELMAP_H
#define DAQCHANNELMAP_H
#include "DAQChannelMap/DAQChannelMapConstants.h"

namespace daqchannelmap
{
  class DAQChannelMap
  {
  public:
    
    static DAQChannelMap* getInstance(int detID);
    
    ~DAQChannelMap();
    
    //==================================================================
    // The functions in this block are common to all detectors
    
    ///
    /// Encode DAQ format information into a single 32 bit identifier.
    ///
    /// \param detID   - detector ID from NovaDAQConventions
    /// \param diblock - diblock ID number. Starts from 1
    /// \param dcm     - DCM id within diblock, starts from 1
    /// \param feb     - FEB id within DCM, starts from 0
    /// \param pixel   - pixel id within FEB, starts from 0
    /// \returns 32-bit encoded channel "DAQ" ID
    ///
    dchan encodeDChan(int       detID,
		      diblock_t diblock,
		      dcm_id_t  dcm,
		      feb_t     feb,
		      pixel_t   pixel) const;
    ///
    /// Encode logical format information (physical detector
    /// coordinates) into a single 32 bit identifier.
    ///
    /// \param detID - detector ID from NovaDAQConventions
    /// \param plane - global plane number, starts from 0 at south
    /// \param cell  - global cell number, starts from 0 at bottom
    /// \returns 32-bit encoded channel "logical" ID
    ///
    lchan encodeLChan(int detId, plane_t plane, cell_t cell) const;
    
    /// Return the detector type given either a dchan or an lchan
    /// identifier - in this case the decoding is the same
    int getDetector(uint32_t anychan) const;
    
    /// Decode the diblock ID from a dchan
    diblock_t getDiBlock(dchan daqchan) const;
    
    /// Decode the dcm ID from a dchan
    dcm_id_t getDCM(dchan daqchan) const;
    
    /// Decode the feb id from a dchan
    feb_t getFEB(dchan daqchan) const;
    
    /// Decode the pixel id from a dchan
    pixel_t getPixel(dchan daqchan) const;
    
    /// Decode the plane number from an lchan
    plane_t getPlane(lchan logicalchan) const;
    
    /// Decode the cell number from an lchan
    cell_t getCell(lchan logicalchan) const;
    //
    //==================================================================

    //==================================================================
    // The functions in this block must be implemented by the
    // subclasses as they are detector-specific

    /// Convert a logical channel ID to a DAQ channel id
    virtual dchan encodeDChan(lchan logicalchan) const = 0;
    
    /// Convert a DAQ channel ID to a logical channel id
    virtual lchan encodeLChan(dchan daqchan) const = 0;

    /// Which view (X or Y) does this daq channel readout?
    virtual DetView_TYPE getView(dchan daqchan) const = 0;

    /// Which block is this lchan in?
    virtual block_t computeBlock(lchan logicalchan) const = 0;

    /// Which global plane in this lchan in?
    virtual plane_t computePlaneInBlock(lchan logicalchan) const = 0;
    
    /// Which module in this lchan in?
    virtual unsigned int computeModule(lchan logicalchan) const = 0;
    
    /// Which fiber does this dchan readout?
    virtual fiber_t computeFiberInModule(dchan daqchan) const = 0;

    /// How many diblocks does the detector have?
    virtual unsigned int getTotalNumberOfDiblocks() const = 0;
    
    /// How many DCMs does this diblock have?
    virtual unsigned int getNumberOfDCM(DiBlock_TYPE dbt) const = 0;

    /// What format is the diblock? Only relevant in NDOS.
    virtual DiBlock_TYPE getDiBlockType(dchan chan) const = 0;
    
    //
    //================================================================== 
  protected:
    /// DAQChannelMap is a singleton - keep the constructor private
    DAQChannelMap();
  };
}

#endif
