#ifndef _DBUTILS_DAQCONFIG_NAMEDCONFIGUTILS_H
#define _DBUTILS_DAQCONFIG_NAMEDCONFIGUTILS_H

#include <DatabaseUtils/DAQConfig/IDSpec.h>
#include <DatabaseUtils/DAQConfig/GlobalNameSpec.h>
#include <DatabaseUtils/DAQConfig/SubsystemNameSpec.h>
#include <DatabaseUtils/SimpleParameterSet.h>
#include <NovaDatabase/Table.h>
#include <NovaDAQUtilities/Status.h>

#include <boost/shared_ptr.hpp>
#include <string>
#include <vector>

namespace dbutils {
namespace daqconfig {

/**
 * Utilities for accessing DAQ named configurations in the database.
 *
 * @author Kurt Biery
 * @version $Revision: 1.31 $ $Date: 2013/02/11 15:36:10 $
 */

class NamedConfigUtils {

 public:

    static const std::string DEFAULT_GLOBAL_NAMES_TABLE_FILE;
    static const std::string DEFAULT_SUBSYSTEM_NAMES_TABLE_FILE;
    static const std::string DEFAULT_DCS_GLOBAL_NAMES_TABLE_FILE;
    static const std::string DEFAULT_DCS_SUBSYSTEM_NAMES_TABLE_FILE;

    static const std::string ID_COLUMN_NAME;
    static const std::string NAME_COLUMN_NAME;
    static const std::string DCMOPMODE_COLUMN_NAME;
    static const std::string DETECTORID_COLUMN_NAME;  // deprecated
    static const std::string DETECTORNAME_COLUMN_NAME;
    static const std::string DESCRIPTION_COLUMN_NAME;
    static const std::string SUBSYSTEMNAME_COLUMN_NAME;
    static const std::string CONFIGTYPE_COLUMN_NAME;
    static const std::string PARENTID_COLUMN_NAME;
    static const std::string GLOBALCFGID_COLUMN_NAME;
    static const std::string SUBSYSCFGID_COLUMN_NAME;
    static const std::string USAGECOUNT_COLUMN_NAME;
    static const std::string ISOBSOLETE_COLUMN_NAME;
    static const std::string ISCHANGING_COLUMN_NAME;
    static const std::string ISCOMPLETE_COLUMN_NAME;
    static const std::string INSERTTIME_COLUMN_NAME;
    static const std::string INSERTUSER_COLUMN_NAME;
    static const std::string UPDATETIME_COLUMN_NAME;
    static const std::string UPDATEUSER_COLUMN_NAME;

    static const std::string NDOS_DETECTOR_NAME;
    static const std::string NEAR_DETECTOR_NAME;
    static const std::string FAR_DETECTOR_NAME;

    static const std::string LATEST_TIME_STRING;

    static const int32_t COPY_SUCCEEDED;
    static const int32_t COPY_FAILED;
    static const int32_t COPY_FOUND_UPDATE_IN_PROGRESS;
    static const int32_t COPY_FOUND_INTERMEDIATE_UPDATE;


    typedef std::map<std::string, std::string> NamedConfigDetails;
    typedef boost::shared_ptr<NamedConfigDetails> NamedConfigDetailsPtr;

    typedef std::map<std::string, std::string> NamedConfigSelectionCriteria;

    typedef boost::shared_ptr<GlobalNameSpec> GlobalNameSpecPtr;
    typedef std::vector<GlobalNameSpec> GlobalNameSpecList;
    typedef std::map<GlobalNameSpec, std::vector<GlobalNameSpec> >
        GlobalNameSpecMap;
    typedef std::map<SubsystemNameSpec, std::vector<SubsystemNameSpec> >
        SubsystemNameSpecMap;


    static bool
        getGlobalConfigId(const GlobalNameSpec& globalNameSpec,
                          int64_t& globalConfigId,
                          const std::string& globalNamesTableFile =
                          DEFAULT_GLOBAL_NAMES_TABLE_FILE);

    static bool
        getSubsystemConfigId(const SubsystemNameSpec& subsystemNameSpec,
                             int64_t& subsystemConfigId,
                             const std::string& subsystemNamesTableFile =
                             DEFAULT_SUBSYSTEM_NAMES_TABLE_FILE);

    static bool getConfigIdAncestry(const IDSpec& idSpec,
                                    std::vector<int64_t>& idList,
                                    const std::string& tableFile =
                                    DEFAULT_GLOBAL_NAMES_TABLE_FILE);

    static bool
        getGlobalConfigIdAncestry(const GlobalNameSpec& globalNameSpec,
                                  std::vector<int64_t>& idList,
                                  const std::string& globalNamesTableFile =
                                  DEFAULT_GLOBAL_NAMES_TABLE_FILE);

    static bool
        getSubsystemConfigIdAncestry(const SubsystemNameSpec&
                                     subsystemNameSpec,
                                     std::vector<int64_t>& idList,
                                     const std::string&
                                     subsystemNamesTableFile =
                                     DEFAULT_SUBSYSTEM_NAMES_TABLE_FILE);

    static bool
        getSubsystemConfigIdFromGlobalName(const GlobalNameSpec&
                                           globalNameSpec,
                                           const std::string& subsystemName,
                                           const std::string& subsystemConfigType,
                                           int64_t& subsystemConfigId,
                                           const std::string&
                                           globalNamesTableFile =
                                           DEFAULT_GLOBAL_NAMES_TABLE_FILE,
                                           const std::string&
                                           subsystemNamesTableFile =
                                           DEFAULT_SUBSYSTEM_NAMES_TABLE_FILE);

    static bool
        getSubsystemConfigIdFromGlobalId(const IDSpec& idSpec,
                                         const std::string& subsystemName,
                                         const std::string& subsystemConfigType,
                                         int64_t& subsystemConfigId,
                                         const std::string&
                                         globalNamesTableFile =
                                         DEFAULT_GLOBAL_NAMES_TABLE_FILE,
                                         const std::string&
                                         subsystemNamesTableFile =
                                         DEFAULT_SUBSYSTEM_NAMES_TABLE_FILE);

    static bool
        getSubsystemConfigIdAncestryFromGlobalId(const IDSpec& globalConfigId,
                                                 const std::string& subsystemName,
                                                 const std::string& subsystemConfigType,
                                                 std::vector<int64_t>& idList,
                                                 const std::string&
                                                 globalNamesTableFile =
                                                 DEFAULT_GLOBAL_NAMES_TABLE_FILE,
                                                 const std::string&
                                                 subsystemNamesTableFile =
                                                 DEFAULT_SUBSYSTEM_NAMES_TABLE_FILE);

    static bool
        getSubsystemConfigIdsFromGlobalId(const IDSpec& idSpec,
                                          std::vector<int64_t>& idList,
                                          std::vector<SubsystemNameSpec>&
                                          subsystemNameSpecList,
                                          const std::string&
                                          globalNamesTableFile =
                                          DEFAULT_GLOBAL_NAMES_TABLE_FILE,
                                          const std::string&
                                          subsystemNamesTableFile =
                                          DEFAULT_SUBSYSTEM_NAMES_TABLE_FILE);

    static std::vector<NamedConfigDetailsPtr>
        getSubsystemConfigsFromGlobalId(const IDSpec& idSpec,
                                        const std::string&
                                        globalNamesTableFile =
                                        DEFAULT_GLOBAL_NAMES_TABLE_FILE,
                                        const std::string&
                                        subsystemNamesTableFile =
                                        DEFAULT_SUBSYSTEM_NAMES_TABLE_FILE);

    static NamedConfigDetailsPtr
        getNewestGlobalConfigFromSubsystemName(const SubsystemNameSpec&
                                               subsystemNameSpec,
                                               const std::string&
                                               subsystemNamesTableFile =
                                               DEFAULT_SUBSYSTEM_NAMES_TABLE_FILE,
                                               const std::string&
                                               globalNamesTableFile =
                                               DEFAULT_GLOBAL_NAMES_TABLE_FILE);

    static NamedConfigDetailsPtr
        getNewestGlobalConfigFromSubsystemId(const IDSpec& idSpec,
                                             const std::string&
                                             subsystemNamesTableFile =
                                             DEFAULT_SUBSYSTEM_NAMES_TABLE_FILE,
                                             const std::string&
                                             globalNamesTableFile =
                                             DEFAULT_GLOBAL_NAMES_TABLE_FILE);

    static NamedConfigDetailsPtr
        getNewestGlobalConfigFromGlobalId(const IDSpec& idSpec,
                                          const std::string&
                                          globalNamesTableFile =
                                          DEFAULT_GLOBAL_NAMES_TABLE_FILE);

    static GlobalNameSpecPtr
        getNewestInstanceOfLatestUsedDAQGlobalConfig(const std::string& detectorName,
                                                     const int& partitionNumber,
                                                     const std::string& globalNamesTableFile =
                                                     DEFAULT_GLOBAL_NAMES_TABLE_FILE);

    static bool
        getRecentGlobalConfigs(IDSpec::MAJOR_SYSTEM_TYPE daqOrSlowControls,
                               const std::string& detectorName,
                               int resultOffset, int resultLimit,
                               GlobalNameSpecList& resultList,
                               const std::string& globalNamesTableFile =
                               DEFAULT_GLOBAL_NAMES_TABLE_FILE);

    static bool
        getAvailableGlobalConfigs(IDSpec::MAJOR_SYSTEM_TYPE daqOrSlowControls,
                                  const std::string& detectorName,
                                  GlobalNameSpecMap& resultsTable,
                                  int maxDetailCountPerSummary = 10,
                                  const std::string& globalNamesTableFile =
                                  DEFAULT_GLOBAL_NAMES_TABLE_FILE);

    static bool
        getAvailableSubsystemConfigs(IDSpec::MAJOR_SYSTEM_TYPE daqOrSlowControls,
                                     const std::string& detectorName,
                                     const std::string& subsystemName,
                                     const std::string& configType,
                                     SubsystemNameSpecMap& resultsTable,
                                     int maxDetailCountPerSummary = 10,
                                     const std::string&
                                     subsystemNamesTableFile =
                                     DEFAULT_SUBSYSTEM_NAMES_TABLE_FILE);

    static novadaq::Status
        copyGlobalConfig(const IDSpec& sourceId, bool makeParentChildLink,
                         const SimpleParameterSet& valuesToOverride,
                         std::string baselineTimestamp,
                         bool overrideIntermediateUpdate,
                         int64_t& resultId,
                         const std::string& globalNamesTableFile =
                         DEFAULT_GLOBAL_NAMES_TABLE_FILE);

    static novadaq::Status
        copySubsystemConfig(const IDSpec& sourceId, bool makeParentChildLink,
                            const SimpleParameterSet& valuesToOverride,
                            std::string baselineTimestamp,
                            bool overrideIntermediateUpdate,
                            int64_t& resultId,
                            const std::string& subsystemNamesTableFile =
                            DEFAULT_SUBSYSTEM_NAMES_TABLE_FILE);

    static bool markConfigDone(const IDSpec& idSpec,
                               bool alsoComplete = false,
                               const std::string& tableFile =
                               DEFAULT_GLOBAL_NAMES_TABLE_FILE);

    static bool incrementUsageCount(const IDSpec& idSpec,
                                    const std::string& tableFile =
                                    DEFAULT_GLOBAL_NAMES_TABLE_FILE);

    static std::vector<NamedConfigDetailsPtr>
        getGlobalConfigDetailAncestry(const GlobalNameSpec&
                                      globalNameSpec,
                                      const std::string&
                                      globalNamesTableFile =
                                      DEFAULT_GLOBAL_NAMES_TABLE_FILE);

    static std::vector<NamedConfigDetailsPtr>
        getConfigDetailAncestry(const IDSpec& idSpec,
                                const std::string& tableFile =
                                DEFAULT_GLOBAL_NAMES_TABLE_FILE);

    static bool
        getFieldValue(const NamedConfigDetailsPtr& details,
                      const std::string& fieldName,
                      std::string& fieldValue);

 private:

    static NamedConfigDetailsPtr
        _getDetails(const boost::shared_ptr<nova::database::Table>& dbt,
                    const NamedConfigSelectionCriteria& selections);

    static bool
        _getId(const boost::shared_ptr<nova::database::Table>& dbt,
               const NamedConfigSelectionCriteria& selections,
               int64_t& id);

    static bool
        _findNextAncestor(const boost::shared_ptr<nova::database::Table>& dbt,
                          std::vector<int64_t>& idList);

    static bool
        _findNextChild(const boost::shared_ptr<nova::database::Table>& dbt,
                       std::vector<int64_t>& idList);
};

};  // end of namespace daqconfig
};  // end of namespace dbutils

#endif
