#include <cstdio>
#include <cstdlib>
#include <iostream>
#include "DAQChannelMap/FarDetDAQChannelMap.h"
#include "DAQChannelMap/NDOSDAQChannelMap.h"
#include "DAQChannelMap/NearDetDAQChannelMap.h"
#include "DAQChannelMap/BitFields.h"
#include "DAQChannelMap/boundscheck.h"
#include "NovaDAQConventions/DAQConventions.h"
using namespace daqchannelmap;

// Create an initial detector specific instance if there is no current
// instance or return the current instance.
//
DAQChannelMap* DAQChannelMap::getInstance(int detID)
{
  static FarDetDAQChannelMap*   FarDetInst = 0;
  static NearDetDAQChannelMap* NearDetInst = 0;
  static NDOSDAQChannelMap*       NDOSInst = 0;

  switch(detID) {
  case novadaq::cnv::kFARDET :
    if (NearDetInst) { delete NearDetInst; NearDetInst = 0; }
    if (NDOSInst)    { delete NDOSInst;    NDOSInst    = 0; }
    if (FarDetInst==0) FarDetInst = new FarDetDAQChannelMap();
    return FarDetInst;
  case novadaq::cnv::kNEARDET :
    if (FarDetInst)  { delete FarDetInst;  FarDetInst = 0; }
    if (NDOSInst)    { delete NDOSInst;    NDOSInst   = 0; }
    if (NearDetInst==0) NearDetInst = new NearDetDAQChannelMap();
    return NearDetInst;
  case novadaq::cnv::kNDOS :
    if (FarDetInst)  { delete FarDetInst;  FarDetInst  = 0; }
    if (NearDetInst) { delete NearDetInst; NearDetInst = 0; }
    if (NDOSInst==0) NDOSInst = new NDOSDAQChannelMap();
    return NDOSInst;
  default:
    std::cerr<<" No valid detector specified!!!!!"<<std::endl;
    abort();
  }
  return 0;
}

//......................................................................

DAQChannelMap::DAQChannelMap() { };

//......................................................................

DAQChannelMap::~DAQChannelMap() { };

//......................................................................

dchan DAQChannelMap::encodeDChan(int       detector,
				 diblock_t diblock,
				 dcm_id_t  dcm,
				 feb_t     feb, 
				 pixel_t   pixel) const
{
  dchan chan = 0x0;
  BITFIELDSET(chan,detector, MASK_DETECTOR, SHIFT_DETECTOR);
  BITFIELDSET(chan,diblock,  MASK_DIBLOCK,  SHIFT_DIBLOCK);
  BITFIELDSET(chan,dcm,      MASK_DCMID,    SHIFT_DCMID);
  BITFIELDSET(chan,feb,      MASK_DCMPORT,  SHIFT_DCMPORT);
  BITFIELDSET(chan,pixel,    MASK_PIXELID,  SHIFT_PIXELID);
  return chan;
}

//......................................................................

lchan DAQChannelMap::encodeLChan(int     detector,
				 plane_t plane,
				 cell_t  cell) const
{
  lchan chan = 0x0;
  BITFIELDSET(chan,detector, MASK_OFFLINE_DETECTOR, SHIFT_OFFLINE_DETECTOR);
  BITFIELDSET(chan,plane , MASK_OFFLINE_PLANE   , SHIFT_OFFLINE_PLANE);
  BITFIELDSET(chan,cell  , MASK_OFFLINE_CELL    , SHIFT_OFFLINE_CELL);
  return chan;
}

//......................................................................

int DAQChannelMap::getDetector(dchan daqchan) const 
{
  return (int)BOUNDSCHECK(BITFIELDGET(daqchan,MASK_DETECTOR,SHIFT_DETECTOR),
			  novadaq::cnv::kUNKNOWN_DET+1,
			  novadaq::cnv::kNDetector-1,
			  ERROR_UNKNOWN_DET);
}

//......................................................................

diblock_t DAQChannelMap::getDiBlock(dchan daqchan) const 
{
  return BOUNDSCHECK(BITFIELDGET(daqchan,MASK_DIBLOCK,SHIFT_DIBLOCK),
		     MIN_DIBLOCK,
		     MAX_DIBLOCK,
		     ERROR_DIBLOCK_OUTRANGE);
}

//......................................................................

dcm_id_t DAQChannelMap::getDCM(dchan daqchan) const
{
  return  BOUNDSCHECK(BITFIELDGET(daqchan,MASK_DCMID,SHIFT_DCMID),
		      MIN_DCM,
		      MAX_DCM,
		      ERROR_DCM_OUTRANGE);
}

//......................................................................

feb_t DAQChannelMap::getFEB(dchan daqchan) const
{
  return BOUNDSCHECK(BITFIELDGET(daqchan,MASK_DCMPORT,SHIFT_DCMPORT),
		     MIN_PORT,
		     MAX_PORT,
		     ERROR_PORT_OUTRANGE);
}

//......................................................................

pixel_t DAQChannelMap::getPixel(dchan daqchan) const
{
  return BOUNDSCHECK(BITFIELDGET(daqchan,MASK_PIXELID,SHIFT_PIXELID),
		     MIN_PIXEL,
		     MAX_PIXEL,
		     ERROR_PIXEL_OUTRANGE);
}

//......................................................................

plane_t DAQChannelMap::getPlane(lchan logicalchan) const
{
  return BOUNDSCHECK(BITFIELDGET(logicalchan, MASK_OFFLINE_PLANE, SHIFT_OFFLINE_PLANE),
		     0,
		     NUM_PLANES_DIBLOCK[getDetector(chan)]*
		     NUM_DIBLOCK_DET[getDetector(chan)], ERROR_PLANE_OUTRANGE);
}

//......................................................................

cell_t DAQChannelMap::getCell(lchan logicalchan) const 
{
  return BOUNDSCHECK(BITFIELDGET(logicalchan,MASK_OFFLINE_CELL,SHIFT_OFFLINE_CELL),
		     0,
		     NUM_CELLS*
		     NUM_PLANES_DIBLOCK[getDetectorOffline(chan)]*
		     NUM_DIBLOCK_DET[getDetectorOffline(chan)],
		     ERROR_CELL_OUTRANGE);
}
////////////////////////////////////////////////////////////////




  
