#include <iostream>
#include <iomanip>
#include <vector>
#include <cstdlib>
#include <ctime>
#include <getopt.h>
#include <string>
#include <fstream>
#include <boost/tokenizer.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/algorithm/string/case_conv.hpp>

#include "NovaDAQUtilities/LocationUtils.h"
#include "NovaDatabase/Table.h"

using namespace nova::database;
using namespace std;

string dbFname = "";
string modFile = "";
bool   setProduction = false;

void PrintUsage()
{
  cout << "Usage: modifyTableInDB [table xml file] [modification file]" 
       << endl;
  cout << "options:\n";
  cout << "\t -h (--help): Print this message" << endl;
  cout << "\t NOTE: the modification table should be of the format:\n";
  cout << "\t [primary key(s)] [columun to modify] [new value]" << endl;
}

//------------------------------------------------------------
bool ParseCLArgs(int argc, char* argv[])
{
  if (argc < 3) {
    PrintUsage();
    exit(0);
  }

  if (argc == 3) {
    dbFname = argv[1];
    modFile = argv[2];
    return true;
  }

  struct option long_options[] = {
    {"help",   0, 0, 'h'},
    {0,0,0,0}
  };
  
  while (1) {
    int optindx;
    int c = getopt_long(argc,argv,"h",long_options,&optindx);
        
    if (c==-1) break;
    
    switch(c) {
    case 'h':
      PrintUsage();
      exit(0);
    default:
      break;
    }
    
  }
  
  if ((argc-optind)==2) {
    dbFname = argv[optind-1];
    modFile = argv[optind];
  }
  else {
    PrintUsage();
    exit(0);    
  }
  
  return true;
}

//------------------------------------------------------------

int main(int argc, char *argv[])
{
  ParseCLArgs(argc, argv);

  Table* dbt;

  try {
    dbt = new Table(dbFname);
  }
  catch(std::runtime_error& e) {
    std::cerr << e.what() << "  Exiting..." << std::endl;
    return 1;
  }

  if (! dbt->GetConnection()) {
    std::cerr << "Could not connect to the database.  Exiting..." << std::endl;
    return 1;
  }

  if (! dbt->ExistsInDB()) {
    std::cerr << dbt->Name() << " does not exist in the database." 
	      << std::endl;
    return 0;
  }

  dbt->LoadFromDB();

  int nPKey = dbt->GetNPKeyCol();
  vector<string> pKeyName;
  for (int i=0; i<nPKey; ++i) 
    pKeyName.push_back(dbt->GetPKeyCol(i)->Name());

  ifstream fin;
  fin.open(modFile.c_str());
  if (!fin.is_open() || !fin.good()) {
    cerr << "Could not open " << modFile << endl;
    return false;
  }
  
  using namespace boost;

  string s;
  while (!fin.eof()) {
    getline(fin,s);
    if (fin.eof()) break;
 
    tokenizer<escaped_list_separator<char> >  tok(s);
    tokenizer<escaped_list_separator<char> >::iterator itr=tok.begin();
    tokenizer<escaped_list_separator<char> >::iterator itrEnd=tok.end();

    vector<string> pkeyVal;

    int i=0;
    for (; itr != itrEnd && i < nPKey; ++itr,++i) {
      string str = *itr;
      boost::trim(str);
      pkeyVal.push_back(str);
    }

    if ((int)pkeyVal.size() != nPKey) {
      std::cerr << "Supplied " << pkeyVal.size() << " primary key values, ";
      std::cerr << "table requires " << dbt->GetNPKeyCol() << endl;      
      continue;
    }

    vector<string> colName;
    vector<string> colValue;

    while (itr != itrEnd) {
      string str = *itr++;
      boost::trim(str);
      colName.push_back(str);
      str = *itr++;
      boost::trim(str);
      colValue.push_back(str);
    }
    
    // now look up row in table
    Row* row = dbt->GetRowByPKey(pkeyVal);

    if (! row) {
      row = dbt->NewRow();
      // set primary key values
      for (unsigned int j=0; j<pkeyVal.size(); ++j) 
	row->Set(pKeyName[j],pkeyVal[j]);
      for (unsigned int j=0; j<colName.size(); ++j)
	row->Set(colName[j],colValue[j]);
      dbt->AddRow(row);
      delete row;
    }
    else {
      for (unsigned int j=0; j<colName.size(); ++j)
	row->Update(colName[j],colValue[j]);
    }

  }
  
  dbt->WriteToDB();
  
  return 0;
}

